# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
import io
import warnings
from typing import Dict, List, Tuple, Union

import pandas as pd
import requests
import yfinance as yf

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

import seaborn as sns; sns.set()
import matplotlib.pyplot as plt


def get_ibovespa(
            data_inicio: str, 
            data_fim: str, 
            proxy: Union[Dict[str, str], None] = None
            ) -> pd.DataFrame:
    if proxy:
        indice_ibov = yf.download("^BVSP", start=data_inicio, end=data_fim, proxy=proxy)
    else:
        indice_ibov = yf.download("^BVSP", start=data_inicio, end=data_fim)
    indice_ibov = indice_ibov[["Adj Close"]]
    retorno_ibov = indice_ibov.pct_change()
    indice_ibov_acumulado = (1 + retorno_ibov).cumprod()
    indice_ibov_acumulado.iloc[0] = 1
    indice_ibov_acumulado.index = pd.to_datetime(indice_ibov_acumulado.index)
    indice_ibov.index = pd.to_datetime(indice_ibov.index)
    return indice_ibov, indice_ibov_acumulado

def get_benchmark(
            data_inicio: str, 
            data_fim: str,
            benchmark: str = "cdi",
            proxy: Union[Dict[str, str], None] = None
            ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if benchmark.upper()=="CDI": codigo_bcb = 12
    elif benchmark.upper()=="IMA-B": codigo_bcb = 12466
    elif benchmark.upper()=="IMA-B 5": codigo_bcb = 12467
    elif benchmark.upper()=="IMA-B 5+": codigo_bcb = 12468
    elif benchmark.upper()=="IBOV":
        indice_ibov, indice_ibov_acumulado = get_ibovespa(data_inicio, data_fim, proxy)
        return indice_ibov, indice_ibov_acumulado
    else: raise ValueError("Benchmark não encontrado.")
    url = (
        f"http://api.bcb.gov.br/dados/serie/bcdata.sgs.{codigo_bcb}/dados?formato=json"
    )
    if proxy:
        dados_cvm = requests.get(url, proxies=proxy, verify=False).text
    else:
        dados_cvm = requests.get(url).text
    dados_benchmark = pd.read_json(io.StringIO(dados_cvm))       
    dados_benchmark["data"] = pd.to_datetime(dados_benchmark["data"], dayfirst=True)
    dados_benchmark = dados_benchmark.set_index("data")
    dados_benchmark = dados_benchmark[(dados_benchmark.index>=data_inicio) & (dados_benchmark.index<=data_fim)] / 100
    dados_benchmark.columns = ["CDI"]
    dados_benchmark_acumulado = (1 + dados_benchmark).cumprod()
    dados_benchmark_acumulado.columns = ["CDI"]
    dados_benchmark_acumulado.iloc[0] = 1
    return dados_benchmark, dados_benchmark_acumulado


def get_stocks(
                acoes: Union[List[str], str],
                data_inicio: str,
                data_fim: str,
                proxy: Union[Dict[str, str], None] = None,
                ) -> pd.DataFrame:
    """Função para capturar dados de Ações ou Índices Listados.
    """
    df1 = pd.DataFrame()
    if isinstance(acoes, list):
        for st in acoes:
            if not st.endswith(".SA"): st = st+".SA"
            if proxy:
                df = yf.download(st, start=data_inicio, end=data_fim, proxy=proxy)
            else:
                df = yf.download(st, start=data_inicio, end=data_fim)
            df1 = pd.concat([df, df1], axis=1)
    else:
        if not acoes.endswith(".SA"): acoes = acoes+".SA"
        if proxy:
            df1 = yf.download(acoes, start=data_inicio, end=data_fim, proxy=proxy)
        else:
            df1 = yf.download(acoes, start=data_inicio, end=data_fim)
    df1.index = pd.to_datetime(df1.index)
    return df1


# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
import warnings
from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

import seaborn as sns; sns.set()
import matplotlib.pyplot as plt


def calcula_rentabilidade_fundos(
    dados_fundos_cvm: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    copia = dados_fundos_cvm.copy()
    fundo_acoes_filtrado_transformed = copia.pivot_table(
        index="DT_COMPTC", columns="CNPJ - Nome", values=["VL_QUOTA", "VL_PATRIM_LIQ"]
    )
    fundo_acoes_filtrado_transformed.index = pd.to_datetime(fundo_acoes_filtrado_transformed.index)
    fundo_acoes_filtrado_transformed = fundo_acoes_filtrado_transformed.sort_index()

    cotas_normalizadas = (fundo_acoes_filtrado_transformed["VL_QUOTA"]/ fundo_acoes_filtrado_transformed["VL_QUOTA"].iloc[0])

    rentabilidade_fundos_diaria = fundo_acoes_filtrado_transformed["VL_QUOTA"].pct_change()

    rentabilidade_fundos_acumulada = (1 + rentabilidade_fundos_diaria).cumprod() - 1
    rentabilidade_fundos_total = rentabilidade_fundos_acumulada.iloc[-1].to_frame()
    rentabilidade_fundos_total.columns = ["rentabilidade"]

    rentabilidade_media_anualizada = (rentabilidade_fundos_diaria * (252)).mean(axis=0).dropna().to_frame()
    rentabilidade_media_anualizada.columns = ["rentabilidade"]

    T = fundo_acoes_filtrado_transformed.shape[0]
    retorno_periodo_anualizado = (((fundo_acoes_filtrado_transformed["VL_QUOTA"].iloc[-1]/ fundo_acoes_filtrado_transformed["VL_QUOTA"].iloc[0])** (252 / T)- 1)* 100).dropna().to_frame()
    retorno_periodo_anualizado.columns = ["rentabilidade"]

    rentabilidade_acumulada_por_ano = (rentabilidade_fundos_acumulada.groupby(pd.Grouper(freq="Y")).last(1).T.dropna())
    rentabilidade_acumulada_por_ano.columns = [str(x)[:4] for x in rentabilidade_acumulada_por_ano.columns]

    volatilidade_fundos = rentabilidade_fundos_diaria.std().dropna().to_frame() * np.sqrt(252)
    volatilidade_fundos.columns = ["volatilidade"]

    risco_retorno = pd.concat([volatilidade_fundos, retorno_periodo_anualizado], axis=1)
    return (
            risco_retorno.dropna() * 100,
            cotas_normalizadas * 100,
            rentabilidade_media_anualizada * 100,
            rentabilidade_acumulada_por_ano * 100,
            rentabilidade_fundos_total * 100,
            )

def melhores_e_piores_fundos(
                            df: pd.DataFrame(), num: int = 5
                            ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    return df.nlargest(num, df.columns[0]), df.nsmallest(num, df.columns[0])


def plotar_comparacao_risco_retorno(
                                    df: pd.DataFrame(),
                                    risco_retorno_carteira: Union[Tuple[int, int], None] = None,
                                    risco_retorno_benchmark: Union[Tuple[int, int], None] = None,
                                    nome_carteira: Union[str, None] = None,
                                    nome_benchmark: Union[str, None] = None,
                                    **opcionais: Any
                                    ) -> None:
    if risco_retorno_carteira:
        risco_carteira, retorno_carteira = risco_retorno_carteira
    if risco_retorno_benchmark:
        risco_benchmark, retorno_benchmark = risco_retorno_benchmark

    plt.figure(figsize=(opcionais.get("figsize")))
    sns.scatterplot(
        data=df, y="rentabilidade", x="volatilidade", alpha=0.45, color="gray"
    )
    if risco_retorno_carteira:
        plt.scatter(x=risco_carteira, y=retorno_carteira, marker="o", color="blue", s=200)
    if risco_retorno_benchmark:
        plt.scatter(x=risco_benchmark, y=retorno_benchmark, marker="o", color="red", s=200)
    plt.ylabel("Retorno (%aa)\n", rotation=0, labelpad=-70, loc="top")
    plt.xlabel("Volatilidade (%aa)")
    plt.box(False)
    plt.grid(True, axis="y")
    plt.ylim(opcionais.get("ylim"))
    plt.xlim(opcionais.get("xlim"))
    if risco_retorno_carteira:
        plt.annotate(
            nome_carteira,
            xy=(risco_carteira + 0.5, retorno_carteira),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_carteira"),
            textcoords="offset points",
            color="darkblue",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="blue", connectionstyle="arc3,rad=-0.1"),
        )
    if risco_retorno_benchmark:
        plt.annotate(
            nome_benchmark,
            xy=(risco_benchmark + 0.5, retorno_benchmark),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_benchmark"),
            textcoords="offset points",
            color="darkred",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="r", connectionstyle="arc3,rad=-0.1"),
        )


def plotar_evolucao(
                    df: pd.DataFrame(), lista_fundos: List[str], **opcionais: Any
                    ) -> Union[pd.DataFrame, None]:

    lista_fundos = [x.upper() for x in lista_fundos]
    cnpj = [x for x in df.columns.tolist() if x.split(" // ")[0] in lista_fundos]
    nome = [x for x in df.columns.tolist() if x.split(" // ")[1] in lista_fundos]
    if len(nome) == 0:
        for fundos in df.columns.tolist():
            fd = fundos.split(" // ")[1]
            for fds in lista_fundos:
                if fds in fd:
                    nome.append(fundos)
    colunas = nome + cnpj
    if len(colunas) > 2:
        data = df[colunas].dropna(axis=1, how="all")
        data = data[data > 0].dropna()
        maximo, minimo = (
            data.iloc[-1:].idxmax(axis=1).values[0],
            data.iloc[-1:].idxmin(axis=1).values[0],
        )
        legenda_maximo = maximo.split(" // ")[0]
        legenda_minimo = minimo.split(" // ")[0]
        data.index = pd.to_datetime(data.index)
        plt.figure(figsize=(opcionais.get("figsize")))
        plt.plot(
            data.drop([maximo, minimo], axis=1),
            color=opcionais.get("color"),
            alpha=opcionais.get("alpha"),
        )
        plt.plot(data[maximo], color=opcionais.get("color_maximo"))
        plt.plot(data[minimo], color=opcionais.get("color_minimo"))
        plt.ylabel("Cotas\n", rotation=0, labelpad=-20, loc="top")
        plt.xlabel("")
        plt.box(False)
        plt.grid(True, axis="y")
        plt.ylim(opcionais.get("ylim"))
        plt.xlim(opcionais.get("xlim"))
        plt.annotate(
            legenda_maximo,
            xy=(data.index[-1], data.iloc[-1:].max(axis=1).values[0]),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_maximo"),
            textcoords="offset points",
            color=opcionais.get("color_maximo"),
            weight="bold",
            arrowprops=dict(
                arrowstyle="->",
                color=opcionais.get("color_seta_maximo"),
                connectionstyle="arc3,rad=-0.1",
            ),
        )
        plt.annotate(
            legenda_minimo,
            xy=(data.index[-1], data.iloc[-1:].min(axis=1).values[0]),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_minimo"),
            textcoords="offset points",
            color=opcionais.get("color_minimo"),
            weight="bold",
            arrowprops=dict(
                arrowstyle="->",
                color=opcionais.get("color_seta_minimo"),
                connectionstyle="arc3,rad=-0.1",
            ),
        )
        return data
    elif len(colunas) <=2:
        data = df[colunas].dropna(axis=1, how="all")
        data = data[data > 0].dropna()
        data.index = pd.to_datetime(data.index)
        plt.figure(figsize=(opcionais.get("figsize")))
        plt.plot(
            data,
            color=opcionais.get("color"),
            alpha=opcionais.get("alpha"),
        )
        plt.ylabel("Cotas\n", rotation=0, labelpad=-20, loc="top")
        plt.xlabel("")
        plt.box(False)
        plt.grid(True, axis="y")
        plt.ylim(opcionais.get("ylim"))
        plt.xlim(opcionais.get("xlim"))
        return data
    elif not colunas:
        print("Fundo não encontrado")
        return None


# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
import getpass
import io
import os
import time
import warnings
import zipfile
from datetime import datetime
from typing import List, Union, Dict

import pandas as pd
import requests

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

def get_classes() -> List[str]:
    return [
	        "Fundo de Renda Fixa",
	        "Fundo de Ações",
	        "Fundo Multimercado",
	        "Fundo Cambial",
	        "FMP-FGTS",
	        "FIDC",
	        "FIDC-NP",
	        "FIC FIDC",
	        "FICFIDC-NP",
	        "FIDCFIAGRO",
	        "FII",
	        "FII-FIAGRO",
	        "FIP",
	        "FIP EE",
	        "FIP Multi",
	        "FIC FIP",
	        "FIP CS",
	        "FIP IE",
	        "FIP-FIAGRO",
	        "FUNCINE",
	    	]


def get_fundsregistration(
    classe: Union[List[str], str, None] = None, 
    proxy: Union[Dict[str, str], None] = None) -> pd.DataFrame:
    classes_disponiveis = get_classes()
    start = time.time()
    url = "http://dados.cvm.gov.br/dados/FI/CAD/DADOS/cad_fi.csv"
    if proxy:
        try:
            dados_cvm = requests.get(url, proxies=proxy, verify=False).text
            df = pd.read_csv(io.StringIO(dados_cvm), sep=";", encoding="ISO-8859-1")
        except:
            raise ValueError("Verifique se a proxy está correta. ParserError: Error tokenizing data")
    else:
        try:
            df = pd.read_csv(url, sep=";", encoding="ISO-8859-1")
        except:
            raise ValueError("Informar proxy. HTTPError: authenticationrequired")
    df_filtrado = df[df["SIT"] == "EM FUNCIONAMENTO NORMAL"]
    if classe:
        if not isinstance(classe, list):
            classe = [classe]
        check_classes = [x for x in classe if x not in classes_disponiveis]
        if check_classes:
            raise ValueError(f"Classe não encontrada {check_classes}")
        df_filtrado = df_filtrado[df_filtrado["CLASSE"].isin(classe)]
        print(f"Cadastro finalizado em {round((time.time()-start)/60,2)} minutos")
    return df_filtrado

def _mesclar_bases(cadastro_fundos: pd.DataFrame, informe_diario_fundos: pd.DataFrame) -> pd.DataFrame:
    cadastro_fundos_filtrado = cadastro_fundos[['CNPJ_FUNDO', 'CLASSE', 'DENOM_SOCIAL']]
    dados_completos_filtrados = informe_diario_fundos.merge(cadastro_fundos_filtrado, on=['CNPJ_FUNDO'], how="inner")
    dados_completos_filtrados["CNPJ - Nome"] = dados_completos_filtrados["CNPJ_FUNDO"] + " // " + dados_completos_filtrados["DENOM_SOCIAL"]
    dados_completos_filtrados = dados_completos_filtrados[["CNPJ - Nome", 'DT_COMPTC', 'CLASSE', 'VL_QUOTA', "NR_COTST", "VL_PATRIM_LIQ"]]
    return dados_completos_filtrados

def _get_response(url: str, proxy: Union[Dict[str, str], None] = None) -> pd.DataFrame:
    if proxy:
        dados_cvm = requests.get(url, proxies=proxy, verify=False)
    else:
        dados_cvm = requests.get(url)
    return dados_cvm

def _ler_dados_diarios(ano: int, mes: int, cadastro_fundos: pd.DataFrame, proxy: Union[Dict[str, str], None] = None, num_minimo_cotistas: Union[int, None] = None, patriminio_liquido_minimo: Union[int, None] = None) -> pd.DataFrame:
    arquivo = "inf_diario_fi_{:02d}{:02d}.csv".format(ano, mes)
    url = "http://dados.cvm.gov.br/dados/FI/DOC/INF_DIARIO/DADOS/inf_diario_fi_{:02d}{:02d}.zip".format(ano, mes)
    dados_cvm = _get_response(url, proxy=proxy)
    if str(dados_cvm) != "<Response [200]>":
        url = "http://dados.cvm.gov.br/dados/FI/DOC/INF_DIARIO/DADOS/HIST/inf_diario_fi_{:02d}.zip".format(ano)
        dados_cvm = _get_response(url, proxy=proxy)
    if str(dados_cvm) == '<Response [407]>':
        raise ValueError("Necessário informar proxy correta. Response [407]")
    zf = zipfile.ZipFile(io.BytesIO(dados_cvm.content))
    zf = zf.open(arquivo)
    lines = zf.readlines()
    lines = [i.strip().decode("ISO-8859-1").split(";") for i in lines]
    fundos = pd.DataFrame(lines[1:], columns=lines[0])
    if "TP_FUNDO" in fundos.columns:
        fundos = fundos[fundos["TP_FUNDO"] == "FI"]
    for col in ["VL_PATRIM_LIQ", "NR_COTST", "VL_QUOTA"]:
        fundos[col] = pd.to_numeric(fundos[col])
    if num_minimo_cotistas:
        fundos = fundos[fundos["NR_COTST"] >= num_minimo_cotistas]
    if patriminio_liquido_minimo:
        fundos = fundos[fundos["VL_PATRIM_LIQ"] >= patriminio_liquido_minimo]

    informe_diario_fundos_filtrado = _mesclar_bases(cadastro_fundos, fundos)
    informe_diario_fundos_filtrado = informe_diario_fundos_filtrado.set_index(["DT_COMPTC"])
    informe_diario_fundos_filtrado.index = pd.to_datetime(informe_diario_fundos_filtrado.index)
    return informe_diario_fundos_filtrado
     

def get_brfunds(
			    anos: Union[List[int], int],
			    meses: Union[List[int], int],
                classe: Union[List[str], str, None] = None,
			    num_minimo_cotistas: Union[int, None] = None,
			    patriminio_liquido_minimo: Union[int, None] = None,
			    proxy: Union[Dict[str, str], None] = None,
				) -> pd.DataFrame:
    start = time.time()
    if isinstance(anos, int): anos = [anos]
    else: anos = list(anos)
    if isinstance(meses, int): meses = [meses]
    else: anos = list(anos)
    cadastro_fundos = get_fundsregistration(classe=classe, proxy=proxy)
    
    informe_diario_fundos_historico = pd.DataFrame()
    for ano in anos:
        for mes in meses:
            if ano == datetime.now().year and mes <= datetime.now().month:
                informe_diario_fundos_filtrado = _ler_dados_diarios(ano, mes, cadastro_fundos, proxy, num_minimo_cotistas, patriminio_liquido_minimo)
                informe_diario_fundos_historico = pd.concat([informe_diario_fundos_historico, informe_diario_fundos_filtrado])
            elif ano < datetime.now().year:
                informe_diario_fundos_filtrado = _ler_dados_diarios(ano, mes, cadastro_fundos, proxy, num_minimo_cotistas, patriminio_liquido_minimo)
                informe_diario_fundos_historico = pd.concat([informe_diario_fundos_historico, informe_diario_fundos_filtrado])
        
    print(f"Dados diários finalizados em {round((time.time()-start)/60,2)} minutos")
    return informe_diario_fundos_historico.sort_index()

def get_fidc(ano: int, 
             mes: int, proxy: 
             Union[Dict[str, str], None] = None) -> pd.DataFrame:
    start = time.time()
    url = "http://dados.cvm.gov.br/dados/FIDC/DOC/INF_MENSAL/DADOS/inf_mensal_fidc_{:02d}{:02d}.zip".format(ano, mes)
    if proxy:
        dados_cvm = requests.get(url, proxies=proxy, verify=False)
    else:
        dados_cvm = requests.get(url)
    if str(dados_cvm) == '<Response [404]>':
        raise ValueError("Não há dados para esta data. Response [404]")
    elif str(dados_cvm) == '<Response [407]>':
        raise ValueError("Necessário informar proxy correta. Response [407]")
    arquivo = "inf_mensal_fidc_tab_X_2_{:02d}{:02d}.csv".format(ano, mes)
    zf = zipfile.ZipFile(io.BytesIO(dados_cvm.content))
    zf = zf.open(arquivo)
    lines = zf.readlines()
    lines = [i.strip().decode("ISO-8859-1").split(";") for i in lines]
    end = time.time()
    print(f"Finalizado em {round((end-start)/60,2)} minutos")
    return pd.DataFrame(lines[1:], columns=lines[0])


def get_fip(ano: int, 
            proxy: Union[Dict[str, str], None] = None) -> pd.DataFrame:
    start = time.time()
    url = "http://dados.cvm.gov.br/dados/FIP/DOC/INF_TRIMESTRAL/DADOS/inf_trimestral_fip_{:02d}.csv".format(ano)
    if proxy:
        dados_cvm = requests.get(url, proxies=proxy, verify=False)
    else:
        dados_cvm = requests.get(url)
    if str(dados_cvm) == '<Response [404]>':
        raise ValueError("Não há dados para esta data. Response [404]")
    elif str(dados_cvm) == '<Response [407]>':    
        raise ValueError("Necessário informar proxy correta. Response [407]")
    lines = [i.strip().split(";") for i in dados_cvm.text.split("\n")]
    end = time.time()
    print(f"Finalizado em {round((end-start)/60,2)} minutos")
    return pd.DataFrame(lines[1:], columns=lines[0])



# -*- coding: utf-8 -*-
from setuptools import setup
import subprocess
from os import path

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README_pypi.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(this_directory, 'comparar_fundos_br', 'version.py'), 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.split('"')[1]

packages = \
['comparar_fundos_br']

package_data = \
{'': ['*']}

install_requires = \
['DateTime>=4.5,<5.0',
 'bar-chart-race>=0.1.0,<0.2.0',
 'pandas>=1.4.3,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'yfinance>=0.1.74,<0.2.0']

setup_kwargs = {
    'name': 'comparar-fundos-br',
    'version': version,
    'description': 'Download dados de fundos de investimento e realiza comparações.',
    'long_description': long_description,
    'long_description_content_type': "text/markdown",
    'author': 'Rafael Rodrigues',
    'author_email': 'rafael.rafarod@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': "",
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)